/*
								+----------------------------------+
								|                                  |
								|  ***  GPU image processing  ***  |
								|                                  |
								|   Copyright  -tHE SWINe- 2008   |
								|                                  |
								|           ImageProc.h            |
								|                                  |
								+----------------------------------+
*/

#ifndef __GPU_IMAGE_PROCESSING_INCLUDED
#define __GPU_IMAGE_PROCESSING_INCLUDED

/*
 *	class CGLImageProc
 *		- GPU image processing manager object
 *		- at least one instance must exist while using any of classes below
 */
class CGLImageProc {
protected:
	static int m_n_instance_num;
	static std::vector<CGLFrameBuffer_FBO*> m_frame_buffer_pool;

public:
	/*
	 *	CGLImageProc::CGLImageProc()
	 *		- default constructor
	 */
	CGLImageProc();

	/*
	 *	CGLImageProc::~CGLImageProc()
	 *		- destructor
	 *		- takes care of cleanup once instance counter reaches zero
	 */
	~CGLImageProc();

	/*
	 *	static CGLFrameBuffer_FBO *CGLImageProc::p_GetRenderBuffer(int n_width, int n_height,
	 *		int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_internal_format,
	 *		bool b_depth_buffer, bool b_depth_texture_target, GLenum n_depth_internal_format,
	 *		bool b_stencil_buffer, bool b_stencil_texture_target, GLenum n_stencil_internal_format)
	 *
	 */
	static CGLFrameBuffer_FBO *p_GetRenderBuffer(int n_width, int n_height,
		int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_internal_format,
		bool b_depth_buffer, bool b_depth_texture_target, GLenum n_depth_internal_format,
		bool b_stencil_buffer, bool b_stencil_texture_target, GLenum n_stencil_internal_format);

	/*
	 *	static CGLShaderObject *CGLImageProc::p_CompileLinkReport(CGLState *p_state,
	 *		const TShaderInfo *p_vs, const TShaderInfo *p_fs)
	 *		- compiles and links shaders p_vs and p_fs (one of them can be 0)
	 *		  and returns new CGLShaderObject
	 *		- reports any errors / warnings to stderr
	 *		- if shader couldn't be compiled returns 0
	 */
	static CGLShaderObject *p_CompileLinkReport(CGLState *p_state,
		const TShaderInfo *p_vs, const TShaderInfo *p_fs);

protected:
	static inline void DeleteRenderBuffer(CGLFrameBuffer_FBO *p_buffer);
};

/*
 *	class CGLPerlinNoise
 *		- 2D perlin noise generator on GPU
 *		- original code written by Stefan Gustavson
 */
class CGLPerlinNoise {
protected:
	static const unsigned char m_p_perm[256];
	static const int m_p_grad3[16][3]; // permutation tables

	static const TShaderInfo m_t_fractal_noise; // noise shader source

	CGLTexture_2D *m_p_noise_perm_tex;
	CGLShaderObject *m_p_noise_shader;

	CShaderVectorParameterRef<float> m_t_scale, m_t_turbulence, m_t_amplitude;
	CShaderVectorParameterRef<int> m_t_octave_num;

public:
	/*
	 *	CGLPerlinNoise::CGLPerlinNoise()
	 *		- default constructor; does nothing
	 */
	CGLPerlinNoise();

	/*
	 *	CGLPerlinNoise::~CGLPerlinNoise()
	 *		- default destructor
	 */
	~CGLPerlinNoise();

	/*
	 *	bool CGLPerlinNoise::Render_Noise(CGLState *p_state, CGLTexture_2D *p_dest,
	 *		Vector2f v_phase, float f_scale, float f_amplitude, int n_octave_num,
	 *		float f_persistence)
	 *		- renders turbulence noise to 2D texture p_dest (RGBA are all set to noise value)
	 *		- p_state is noise phase, f_scale is noise size, f_amplitude is amplitude,
	 *		  n_octave_num is number of noise octaves and f_persistence is octave
	 *		  amplitude exponent
	 *		- returns true on success, false on failure
	 */
	bool Render_Noise(CGLState *p_state, CGLTexture_2D *p_dest, Vector2f v_phase, float f_scale,
		float f_amplitude, int n_octave_num, float f_persistence);

protected:
	static unsigned char *p_PermTextureData();
	static CGLTexture_2D *p_MakePermTexture(CGLState *p_state);
};

/*
 *	class CGLGaussFilter
 *		- simple separable gaussian blur filter
 *		- kernel is calculated on GPU, it's not precalculated in texture
 */
class CGLGaussFilter {
protected:
	static const TShaderInfo m_t_separable_gauss; // gauss shader source

	CGLShaderObject *m_p_gauss_shader;

	CShaderVectorParameterRef<float> m_t_org_off, m_t_scale;
	CShaderVectorParameterRef<int> m_t_kernel_size;

	int m_n_radius;
	float m_f_kernel_rescale; // kernel scaling factor cache

public:
	/*
	 *	CGLGaussFilter::CGLGaussFilter()
	 *		- default constructor
	 *		- doesn't do anything
	 */
	CGLGaussFilter();

	/*
	 *	CGLGaussFilter::~CGLGaussFilter()
	 *		- destructor
	 */
	~CGLGaussFilter();

	/*
	 *	bool CGLGaussFilter::Filter(CGLState *p_state, CGLTexture_2D *p_dest,
	 *		CGLTexture_2D *p_temp_tex, CGLTexture_2D *p_src, int n_radius)
	 *		- applies separable gaussian filter with radius n_radius
	 *		- p_src is source image, p_temp_tex is temporary texture and p_dest
	 *		  is destination texture. note all three should have the same resolution.
	 *		- p_state is OpenGL state guard
	 *		- returns true on success, false on failure
	 */
	bool Filter(CGLState *p_state, CGLTexture_2D *p_dest,
		CGLTexture_2D *p_temp_tex, CGLTexture_2D *p_src, int n_radius);
};

/*
 *	class CGLMinMaxFilter
 *		- somewhat nicer separable minmax filter
 *		- finds minima / maxima under cross-shaped kernel
 *		  (cross shape is supposed to be better than ordinary square as
 *		  it is similar to diamond / circle with low radii)
 *		- note this needs support for four draw buffers
 */
class CGLMinMaxFilter {
protected:
	static const TShaderInfo t_minmax_pass1; // gauss shader source
	static const TShaderInfo t_minmax_pass2; // gauss shader source

	CGLShaderObject *m_p_minmax_shader1, *m_p_minmax_shader2;

	CShaderVectorParameterRef<float> m_t_right_off_center, m_t_off_left_off_right;
	CShaderVectorParameterRef<int> m_t_kernel_size_side_size1; // pass 1 params

	CShaderVectorParameterRef<float> m_t_up_off_center, m_t_off_top_off_bottom;
	CShaderVectorParameterRef<int> m_t_kernel_size_side_size2; // pass 2 params

public:
	/*
	 *	CGLMinMaxFilter::CGLMinMaxFilter()
	 *		- default constructor; does nothing
	 */
	CGLMinMaxFilter();

	/*
	 *	CGLMinMaxFilter::~CGLMinMaxFilter()
	 *		- destructor
	 */
	~CGLMinMaxFilter();

	/*
	 *	bool CGLMinMaxFilter::Filter(CGLState *p_state, CGLTexture_2D *p_dest_min,
	 *		CGLTexture_2D *p_dest_max, CGLTexture_2D *p_temp_tex[4], CGLTexture_2D *p_src,
	 *		int n_radius)
	 *		- finds minimal and maximal pixel values in p_src image under cross-shaped
	 *		  kernel with radius n_radius
	 *		- p_dest_min is filled with minima and p_dest_max is filled with maxima
	 *		- p_temp_tex is array of four temporary textures
	 *		- p_state is OpenGL state guard
	 *		- note all textures must have the same dimensions and should have the same format
	 *		- returns true on success, false on failure
	 */
	bool Filter(CGLState *p_state, CGLTexture_2D *p_dest_min, CGLTexture_2D *p_dest_max,
		CGLTexture_2D *p_temp_tex[4], CGLTexture_2D *p_src, int n_radius);
};

#endif //__GPU_IMAGE_PROCESSING_INCLUDED
